from setuptools import setup

setup(
    name='atmos',
    packages=['atmos'],
    version='0.2.1',
    description='Atmospheric sciences utility library',
    author='Jeremy McGibbon',
    author_email='mcgibbon@uw.edu',
    keywords=['atmos', 'atmospheric', 'equations', 'geoscience', 'science'],
    classifiers=[],
)
